Principle of operation
======================

The `mihirestore` image restoration code is intended for the restoration
of spectral datasets that are insufficiently complete for direct use of
the MOMFBD code, or are unsuitable for other reasons. Such datasets
include spectro-graph scan data, and data that carry subtle but
systematic calibration residuals, that would be easily picked up by the
wavefront sensing algorithm of MOMFBD and falsely interpreted as image
features.

Such data can be used to constrain a system of linear equations, that
describe how the data values in the image are composed from many
contributions, each described by the solar intensity value and the value
of the PSF of each object element for a given image element.

If we are able to obtain the PSF $\varphi_{u,v}$ for each point in the
data $d_{x,y}$ by some means, we can describe the contribution of the
spectrum emerging from all points on the Sun $s(u,v)$ to a given data
point $(x,y)$ by

$$d_{x,y} = \sum_{u}^{}{\sum_{v}^{}\varphi_{u,v}}(x - u,y - v)s(u,v),$$

Since each recorded data value can be described by an independent
equation of this type, we have a system of equations that describes the
observed data in terms of known PSF values and a 3D distribution of
unknown solar monochromatic intensities. Every recorded spectral data
frame adds equations to this system, and improves its condition number,
making the soluion more robust (less noisy). If a sufficient number of
linearly independent equations are collected, the system becomes fully
determined and the unknown values of the solar spectra can be obtained.
This proces is described in more detail in (van Noort 2017).

The `mihirestore` code was developed to set up this system of equations,
by combining the wavefront sensing information obtained from a wide-band
reference camera, using a variety of suitable simultaneously recorded
observational data. To accomplish this task, it is a prerequisite that
there is a reliable and accurately known relation in time between the
data channels.

Once the equations have been formulated and consolidated into a Hessian,
the code proceeds to obtain the solution of the system of equations by
means of a variation on the Lucy-Richardson deconvolution algorithm, and
uses Orthomin acceleration to increase the speed of convergence.

Data preparation
================

Wavefront sensing
-----------------

The first step in the data preparation is to obtain the PSFs that are
needed by the `mihirestore` code to construct the system of linear
equations. These PSFs are usually obtained from a context camera,
strictly synchronized to the spectral cameras, and result from a fit to
the data using a process known as Multi-Frame Blind Deconvolution
(MFBD), in which the PSF is assumed to result solely from deviations of
the wavefront from flat, caused by refractive index variations along the
line of sight (LOS). Under the assumption that the object does not
change appreciably during the acquisition of the reference images, a
joint estimate of the individual PSFs and the undegraded object can be
made, using an iterative optimization algorithm (for details see for
instance van Noort, Rouppe van der Voort, and Löfdahl (2005)). The
`mihirestore` code currently only supports restored data files generated
using a recent version of the original MOMFBD code.

In the present context, the restored images are not of great interest,
and are used only to provide alignment information. Instead, we are
after the fitted wavefront coefficients, from which the atmospheric PSF
can be calculated. The data must therefore be saved in the MOMFBD native
format, with the keywords `GET_ALPHA`, `GET_MODES` and `GET_DIVERSITY`
enabled (this is required with or without a phase diversity channel), so
that the PSF can be constructed from it. Because the image scale of the
spectral data can differ significantly from that of the context data,
this is the preferred (and currently only) form of providing the PSF to
the `mihirestore` code, because it allows for easy scaling of the PSF to
match the image scale of the spectral data.

Destretching
------------

Once the reference data have been restored, the first step of
post-processing is carried out. For that to be useful, the data must
consist of a continuous time series, without significant drops in the
seeing quality, and of sufficient length. The time series is reduced
with significant overlap in time, so that each frame is included in at
least two restorations. The assumption is then made that the tip-tilt
coefficient of any given patch in a given frame is really the same, and
that the difference in the value it has in one restoration compared to
another restoration (using different companion frames) is caused by an
offset between the frames of reference of both restorations, caused by
the incorrect assumption that the average of the tip-tilt over the
restoration vanishes. By averaging the difference in the tip-tilt
coefficients over many frames, we can thus recover the offset between
the restored images with a fairly high accuracy.

Although the assumption that the tip-tilt vanishes is incorrect on the
time scale of one restoration, which is typically close to the solar
evolution time scale, is is quite accurate over much longer time scales.
We can therefore average all the displacements of the restored frames
over the full length of a time series, and use that as the reference
coordinates. The position of every restorede patch and frame can now be
calculated relative to that one reference point, thus allowing each
frame to be restored in the same \"absolute\" coordinate system.

Usually the large scale evolution of the Sun or the rotation of the
earth cause a slow drift of the pointing, which may cause the FOV to
move considerably in absolute coordinates. To avoid this, instead of
restoring the frames in one coordinate system, a slowly changing
coordinate system is used instead, which can be obtained by retaining
only the high frequency components of the frame-to-frame offsets, but
allowing for a smooth, slow drift.

The code `wfc_align` was written to calculate the offsets needed for
this task, and write out a file with "absolute" coordinates for each
frame it is given.

Dark and gain corrections
-------------------------

To prepare the spectral data for processing, it must first be corrected
for dark current, and any gain variations that might exist in the data.
These data must then be stored in a separate directory, and carry the
same name as the first channel of the simultaneously recorded images
that are used for the wavefront sensing (the context imager).

Alignment and scale
-------------------

To formulate the linear system of equations, the PSF of the reference
data must be accurately aligned with the spectral data, and scaled
aprropriately. This can be difficult, since the spectral data usually do
not contain much spatial information in at least one direction, which
makes it difficult to compare them with the reference images. For
spectral scans, even a simple summation of the data to compose a scan of
raw data can be sufficient for this task, but if the scan is slow, it
can be difficult to identify what is in reality a diagonal slice out of
the x,y,$\lambda$ domain with an instantaneous image.

Running the code
================

The Mihirestore package was written in C++ for a standard Linux platform
with no special dependcies other than the FFTW3 library. Both
sub-programs (`wfc_align` and `mihirestore`) are compiled with a static
Makefile.

wfc\_align
----------

Usage: `wfc_align [options] FILE`

`FILE` must be a ASCII text file containing a sorted and uninterrupted
list of MOMFBD files, that overlap in time. Each file must be a MOMFBD
format file that contains wavfront coefficients, modes, phase diversity
information.

The code produces the files `sxpos.f0` and `sypos.f0`, which contain the
offset for each patch in the restored files w.r.t. the running average.

### Arguments

-di \#

:   Number of realizations shift between subsequent files

-aw \#

:   The number of restored frames in the window function used to
    calculate the running average of the tip-tilt coefficients. The
    running average is subtracted from the displacements before saving
    them. The default value is "all", effectively subtracting the
    average over the whole time series

-h- -help

:   Some help

-q- -quiet

:   decrease verbosity

-v- -verbose -V

:   increase verbosity

mihirestore
-----------

Usage: `mihirestore [options] FILE`

### Arguments

`FILE` must be a MOMFBD format file that contains wavfront coefficients,
modes, phase diversity information and the list of names of the input
files.

Options
-------

-n- -nthread \#

:   number of threads to use, defaults to 20.

-o- -off \#

:   state in the cycle, defaults to 0

-m- -mod \#

:   length of the cycle after division, defaults to 1

-u- -mul \#

:   Number of context realizations for each spectral realization

-ll- -wlmin \#

:   first wavelength to process.

-lh- -wlmax \#

:   last wavelength to process.

-lr- -lucyrich LIST

:   perform Lucy-Richardson iterations, write each number from the
    provided comma separated list of numbers to disk.

- -lrbase NAME

:   base name for the results.

- -imgbase NAME

:   base name for the destretched MOMFBD image

- -mask NAME

:   Name of wavelength independent mask to apply to the raw data. The
    masked pixels are not included in the system of equations.

- -raw NAME

:   directory containing the raw data files.

-r

:   Read RHS from file

- -rhsname NAME

:   Name of file to save/read the RHS data

-a

:   Read matrix from file

- -lhsname NAME

:   Name of file to save/read the LHS matrix

-dxFLOAT

:   Shift PSF by this amount in the x direction \[in pixels\]

-dyFLOAT

:   Shift PSF by this amount in the y direction \[in pixels\]

-xb \#

:   Height of horizontal border around raw data in x

-yb \#

:   Width of horizontal border around raw data in y

-ndx \#

:   Override data width

-ndy \#

:   Override data height

-xo- -xoffs FLOAT

:   offset between x=0 in the context data and x=0 in the hyperspectral
    data

-yo- -yoffs FLOAT

:   offset between y=0 in the context data and y=0 in the hyperspectral
    data

-sf- -scale FLOAT

:   ratio of context image scale to hyperspectral image scale

-xp- -xpos NAME

:   file containing pixel offsets

-yp- -ypos NAME

:   file containing pixel offsets

-h- -help

:   Some help

-q- -quiet

:   decrease verbosity

-v- -verbose -V

:   increase verbosity

-s- -save

:   save various intermediate results

- -noff

:   disable flatfield correction

- -noslv

:   do not solve the linear system

- -nolhs

:   do not compute the LHS

- -norhs

:   do not compute or read the RHS

References
==========

van Noort, M. 2017. "Image restoration of solar spectra" 608 (December):
A76. <https://doi.org/10.1051/0004-6361/201731339>.

van Noort, Michiel, Luc Rouppe van der Voort, and Mats G. Löfdahl. 2005.
"Solar Image Restoration By Use Of Multi-frame Blind De-convolution With
Multiple Objects And Phase Diversity" 228 (1-2): 191--215.
<https://doi.org/10.1007/s11207-005-5782-z>.
