#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "types.h"
#include "io.h"
#include "mem.h"
#include "uts.h"
//#include "fileio.h"
#include "nr/fft.h"
#include "ana_io.h"
#include "momfbd_read.h"

void patch_struct::read(int fd,void *buf,int sz)
{
  if(::read(fd,buf,sz)<0) fprintf(stderr,"error reading from file: %s\n",strerror(errno));
}

patch_struct::patch_struct(int fd,uint08_t swap_endian,const char *header)
{
  memset(this,0,sizeof(patch_struct));
//
  float32_t version=0.0;
  sscanf(header,"%f",&version);
//
  read(fd,&xl,sizeof(int32_t));
  read(fd,&xh,sizeof(int32_t));
  read(fd,&yl,sizeof(int32_t));
  read(fd,&yh,sizeof(int32_t));
  if(version>=20110714.0){
    read(fd,&offx,sizeof(int32_t));
    read(fd,&offy,sizeof(int32_t));
  }
  read(fd,&nc,sizeof(int32_t));
  if(swap_endian){
    bswap(&xl,1);
    bswap(&xh,1);
    bswap(&yl,1);
    bswap(&yh,1);
    bswap(&nc,1);
  }
//
  nx=xh-xl+1;
  ny=yh-yl+1;
//
  nim=new int32_t [nc]-1;
  dx=new int32_t [nc]-1;
  dy=new int32_t [nc]-1;
  read(fd,nim+1,nc*sizeof(int32_t));
  read(fd,dx+1,nc*sizeof(int32_t));
  read(fd,dy+1,nc*sizeof(int32_t));
  if(swap_endian){
    bswap(nim+1,nc);
    bswap(dx+1,nc);
    bswap(dy+1,nc);
  }
//
  uint08_t has_img;
  read(fd,&has_img,1);
  if(has_img){
    img=f32t2dim(1,ny,1,nx);
    read(fd,img[1]+1,nx*ny*sizeof(float32_t));
    if(swap_endian) bswap(img[1]+1,nx*ny);
  }
//
  read(fd,&npsf,sizeof(int32_t));
  if(swap_endian) bswap(&npsf,1);
  if(npsf){
    psf=f32t3dim(1,npsf,1,ny,1,nx);
    read(fd,psf[1][1]+1,npsf*nx*ny*sizeof(float32_t));
    if(swap_endian) bswap(psf[1][1]+1,npsf*nx*ny);
  }
//
  read(fd,&nobj,sizeof(int32_t));
  if(swap_endian) bswap(&nobj,1);
  if(nobj){
    obj=f32t3dim(1,nobj,1,ny,1,nx);
    read(fd,obj[1][1]+1,nobj*nx*ny*sizeof(float32_t));
    if(swap_endian) bswap(obj[1][1]+1,nobj*nx*ny);
  }
//
  read(fd,&nres,sizeof(int32_t));
  if(swap_endian) bswap(&nres,1);
  if(nres){
    res=f32t3dim(1,nres,1,ny,1,nx);
    read(fd,res[1][1]+1,nres*nx*ny*sizeof(float32_t));
    if(swap_endian) bswap(res[1][1]+1,nres*nx*ny);
  }
//
  read(fd,&nalpha,sizeof(int32_t));
  if(swap_endian) bswap(&nalpha,1);
  if(nalpha){      
    read(fd,&nm,sizeof(int32_t));
    if(swap_endian) bswap(&nm,1);
    alpha=f32t2dim(1,nalpha,1,nm);
    read(fd,alpha[1]+1,nalpha*nm*sizeof(float32_t));
    if(swap_endian) bswap(alpha[1]+1,nalpha*nm);
  }
  if(version>=20100726.0){  // check for version!?
    read(fd,&ndiv,sizeof(int32_t));
    if(swap_endian) bswap(&ndiv,1);
  }
  nphx=nx/2;
  nphy=ny/2;
  if(ndiv){
    if(version>=20110708.0){  // check for version!?
      read(fd,&nphx,sizeof(int32_t));
      if(swap_endian) bswap(&nphx,1);
      read(fd,&nphy,sizeof(int32_t));
      if(swap_endian) bswap(&nphy,1);
    }else fprintf(stderr,"WARNING: version=%.1f, diversity of versions < 20110708.0 may have wrong dimensions!\n",version);
    diversity=f32t3dim(1,ndiv,1,nphy,1,nphx);
    for(int m=1;m<=ndiv;++m){
      int08_t typ;
      if(version>=20110916.0) read(fd,&typ,1);
      read(fd,diversity[m][1]+1,nphx*nphy*sizeof(float32_t));
    }
    if(swap_endian) bswap(diversity[1][1]+1,ndiv*nphx*nphy);
  }
  tfwf=0;
}
  
patch_struct::~patch_struct(void){
  if(tfwf) del_f64t2dim(tfwf,1,nphx,1,nphy);
  if(img) del_f32t2dim(img,1,ny,1,nx);
  if(psf) del_f32t3dim(psf,1,npsf,1,ny,1,nx);
  if(obj) del_f32t3dim(obj,1,nobj,1,ny,1,nx);
  if(res) del_f32t3dim(res,1,nres,1,ny,1,nx);
  if(alpha) del_f32t2dim(alpha,1,nalpha,1,nm);
  if(diversity) del_f32t3dim(diversity,1,ndiv,1,nphy,1,nphx);
  if(nim) delete[] (nim+1);
  if(dx) delete[] (dx+1);
  if(dy) delete[] (dy+1);
}

float64_t **patch_struct::wavefront(int idx,float32_t ***mode,float64_t &dx_io,float64_t &dy_io,float64_t cf2pix)
// return wavefront without tip-tilt, and tip-tilt in dx_io and dy_io
{
  float64_t **wf=f64t2dim(1,nphx,1,nphy);
  memset(wf[1]+1,0,nphx*nphy*sizeof(float64_t));
  float64_t adx=-alpha[idx][1]*cf2pix+sdx+dx_io,ady=-alpha[idx][2]*cf2pix+sdy+dy_io;
  dx_io=adx;
  dy_io=ady;
  for(int m=3;m<=nm;++m) // all modes other than tip-tilt
    for(int u=1;u<=nphx;++u) for(int v=1;v<=nphy;++v) wf[u][v]+=alpha[idx][m]*mode[m][u][v];
  return wf;
}

/*
float64_t **patch_struct::wavefront(int idx,float32_t ***mode,float64_t &dx_io,float64_t &dy_io,float64_t cf2pix)
// return wavefront without tip-tilt, and tip-tilt in dx_io and dy_io
// this doesn't work: if idx changes, the returned wavefront does not!
{
  float64_t **wf=f64t2dim(1,nphx,1,nphy);
  if(!tfwf){
    tfwf=f64t2dim(1,nphx,1,nphy);
    memset(tfwf[1]+1,0,nphx*nphy*sizeof(float64_t));
    for(int m=3;m<=nm;++m) // all modes other than tip-tilt
      for(int u=1;u<=nphx;++u) for(int v=1;v<=nphy;++v) tfwf[u][v]+=alpha[idx][m]*mode[m][u][v];
  }
  float64_t adx=-alpha[idx][1]*cf2pix+sdx+dx_io,ady=-alpha[idx][2]*cf2pix+sdy+dy_io;
  dx_io=adx;
  dy_io=ady;
  memcpy(wf[1]+1,tfwf[1]+1,nphx*nphy*sizeof(float64_t));
  return wf;
}
*/

float64_t **patch_struct::diff(patch_struct *p,int di,int &ncf)
{
  ncf=nim[1]-di; // only compare the reference channel (all other coefficients are the same anyway)
  float64_t **cfs=f64t2dim(1,ncf,1,nm);
  for(int a=1;a<=ncf;++a) for(int m=1;m<=nm;++m) cfs[a][m]=alpha[a][m]-p->alpha[a+di][m];
  return cfs;
}

float32_t patch_struct::imgval(float32_t ix,float32_t iy)
{ // subtract the patch offset from the requested coordinate, to avoid having to add it everywhere else...
  ix-=sdx;
  iy-=sdy;
  if((ix<xl+12)||(ix>=xh-12)||(iy<yl+12)||(iy>=yh-12)) return -1.0;
  int xi=(int)(ix-xl)+1,yi=(int)(iy-yl)+1;
  float32_t fx=ix-(float32_t)(xi+xl-1),fy=iy-(float32_t)(yi+yl-1);
  return fx*(fy*img[xi+1][yi+1]+(1.0-fy)*img[xi+1][yi])+(1.0-fx)*(fy*img[xi][yi+1]+(1.0-fy)*img[xi][yi]);
}


ssize_t momfbd_struct::read(int fd,void *buf,int sz)
{
  ssize_t rv=::read(fd,buf,sz);
  if(rv<0) fprintf(stderr,"error reading from file: %s\n",strerror(errno));
  return rv;
}

momfbd_struct::momfbd_struct(const char *file_name,io_class &io)
{
  strcpy(filename=new char [strlen(file_name)+1],file_name);
  pix2cf=NAN;
  cf2pix=NAN;
  mode=0;
  pupil=0;
  version=0.0;
  date_obs=0;
  time_obs=0;
  nph=0;
  nm=0;
  names=0;
  nsx=0;
  nsy=0;
  np=0;
  nc=0;
  cl_xl=0;
  cl_xh=0;
  cl_yl=0;
  cl_yh=0;
  xo=0.0;
  yo=0.0;
  sf=1.0;
//
  if(int fd=open(file_name,O_RDONLY)){
    int one=1;
    uint08_t endian=(*(char*)&one==0); // 1=LITTLE_ENDIAN, 0=BIG_ENDIAN: taken from SL's tiff library 
    uint08_t f_endian;
    read(fd,&f_endian,1);
    uint08_t swap_endian=(f_endian!=endian);
//
    io.msg(IOL_INFO,"byte swapping: %s\n",(swap_endian)?"yes":"no");
//
    int32_t strsz;
    read(fd,&strsz,sizeof(int32_t));
    if(swap_endian) bswap(&strsz,1);
    char *h=new char [strsz];
    read(fd,h,strsz);
    io.msg(IOL_INFO,"header: %s\n",h);
//
    read(fd,&strsz,sizeof(int32_t));
    if(swap_endian) bswap(&strsz,1);
    read(fd,time_obs=new char [strsz],strsz);
    io.msg(IOL_INFO,"time: %s\n",time_obs);
//
    read(fd,&strsz,sizeof(int32_t));
    if(swap_endian) bswap(&strsz,1);
    read(fd,date_obs=new char [strsz],strsz);
    io.msg(IOL_INFO,"date: %s\n",date_obs);
//
    uint08_t has_modes;
    read(fd,&has_modes,1);
    io.msg(IOL_INFO,"modes: %s\n",(has_modes)?"yes":"no");
    sscanf(h,"%f",&version);
    if(has_modes){
      if(version>=20110714.0){             // check for version!?
        read(fd,&pix2cf,sizeof(float32_t));
        read(fd,&cf2pix,sizeof(float32_t));
        if(swap_endian){
          bswap(&pix2cf,1);
          bswap(&cf2pix,1);
        }
      }
      read(fd,&nph,sizeof(int32_t));
      read(fd,&nm,sizeof(int32_t));
      if(swap_endian){
        bswap(&nph,1);
        bswap(&nm,1);
      }
      pupil=f32t2dim(1,nph,1,nph);
      read(fd,pupil[1]+1,nph*nph*sizeof(float32_t));
      if(swap_endian) bswap(pupil[1]+1,nph*nph);

      io.msg(IOL_INFO,"reading %dx%dx%d ... ",nm,nph,nph);
      mode=f32t3dim(1,nm,1,nph,1,nph);
      for(int m=1;m<=nm;++m){
        int08_t typ;
        if(version>=20110916.0) read(fd,&typ,1); // mode type
        read(fd,mode[m][1]+1,nph*nph*sizeof(float32_t));
      }
//      read(fd,mode[1][1]+1,nm*nph*nph*sizeof(float32_t));
      if(swap_endian) bswap(mode[1][1]+1,nm*nph*nph);
      io.msg(IOL_INFO,"done.\n");
    }
// globals
    read(fd,&nc,sizeof(int32_t));
    if(swap_endian) bswap(&nc,1);
    io.msg(IOL_INFO,"nc=%d\n",nc);
// global clip
    cl_xl=new int16_t [nc];
    cl_xh=new int16_t [nc];
    cl_yl=new int16_t [nc];
    cl_yh=new int16_t [nc];
    if(read(fd,cl_xl,nc*sizeof(int16_t))<nc*sizeof(int16_t)) io.msg(IOL_WARN,"short read: \n",strerror(errno));
    if(read(fd,cl_xh,nc*sizeof(int16_t))<nc*sizeof(int16_t)) io.msg(IOL_WARN,"short read: \n",strerror(errno));
    if(read(fd,cl_yl,nc*sizeof(int16_t))<nc*sizeof(int16_t)) io.msg(IOL_WARN,"short read: \n",strerror(errno));
    if(read(fd,cl_yh,nc*sizeof(int16_t))<nc*sizeof(int16_t)) io.msg(IOL_WARN,"short read: \n",strerror(errno));
    if(swap_endian){
      bswap(cl_xl,nc);
      bswap(cl_xh,nc);
      bswap(cl_yl,nc);
      bswap(cl_yh,nc);
    }
//
    for(int c=0;c<nc;++c) io.msg(IOL_XNFO,"clip=(%d,%d,%d,%d)\n",cl_xl[c],cl_xh[c],cl_yl[c],cl_yh[c]);
//
    read(fd,&nsx,sizeof(int32_t));
    read(fd,&nsy,sizeof(int32_t));
    read(fd,&np,sizeof(int32_t));
    if(swap_endian){
      bswap(&nsx,1);
      bswap(&nsy,1);
      bswap(&np,1);
    }
//
    io.msg(IOL_INFO,"reading %d x %d\n",nsx,nsy);
//
    patch=(patch_struct***)v2dim(1,nsx,1,nsy);
    for(int x=1;x<=nsx;++x)
      for(int y=1;y<=nsy;++y) patch[x][y]=new patch_struct(fd,swap_endian,h);
//
    int32_t n_names=0;
    if(read(fd,&n_names,sizeof(int32_t))<=0) n_names=0;
    if(n_names){
      io.msg(IOL_INFO,"reading %d names\n",n_names);
      names=new char* [n_names+1];
      for(int i=0;i<n_names;++i){
        int32_t ss;
        read(fd,&ss,sizeof(int32_t));
        names[i]=new char [ss];
        read(fd,names[i],ss);
      }
      names[n_names]=0;
    }else io.msg(IOL_INFO,"no names\n");
//
    close(fd);
//
    fft_init(nph,nph);
    fft_init(2*nph,2*nph);
    fft_init(np,np);
  }
}

momfbd_struct::~momfbd_struct(void)
{
  if(sf!=1.0){
    int np_o=2*(int)round(0.5*sf*(float64_t)np);
    fft_done(np_o,np_o);
  }
  fft_done(nph,nph);
  fft_done(2*nph,2*nph);
  fft_done(np,np);
//
  for(int x=1;x<=nsx;++x) for(int y=1;y<=nsy;++y) delete patch[x][y];
  del_v2dim((void***)patch,1,nsx,1,nsy);
  if(cl_xl) delete[] cl_xl;
  if(cl_xh) delete[] cl_xh;
  if(cl_yl) delete[] cl_yl;
  if(cl_yh) delete[] cl_yh;
  if(mode) del_f32t3dim(mode,1,nm,1,nph,1,nph);
  if(pupil) del_f32t2dim(pupil,1,nph,1,nph);
  delete[] date_obs;
  delete[] time_obs;
  for(int n=0;names[n];++n) delete[] names[n];
  delete[] names;
  delete[] filename;
}

void momfbd_struct::map(float64_t xo_in,float64_t yo_in,float64_t sf_in,io_class &io)
{
  if(sf!=1.0){
    int np_o=2*(int)round(0.5*sf*(float64_t)np);
    fft_done(np_o,np_o);
  }
  xo=xo_in;
  yo=yo_in;
  sf=sf_in;
  int np_o=2*(int)round(0.5*sf*(float64_t)np);
  fft_init(np_o,np_o);
  io.msg(IOL_XNFO,"momfbd_struct::map: %.2lf,%.2lf,%.2lf\n",xo,yo,sf);
}

void momfbd_struct::stretch(float64_t **xpos,float64_t **ypos,io_class &io)
{
  if(xpos&&ypos){
    for(int x=1;x<=nsx;++x) for(int y=1;y<=nsy;++y) patch[x][y]->stretch(xpos[x][y],ypos[x][y]);
    io.msg(IOL_XNFO,"momfbd_struct::stretch: stretch maps initialized\n");
  }else io.msg(IOL_XNFO,"momfbd_struct::stretch: stretch maps not initialized\n");
}

void momfbd_struct::info(io_class &io)
{
  io.msg(IOL_DFLT,"file: %s\n",(filename)?filename:"none");
  io.msg(IOL_DFLT,"date: %s\n",(date_obs)?((strlen(date_obs))?date_obs:"n/a"):"n/a");
  io.msg(IOL_DFLT,"time: %s\n",(time_obs)?((strlen(time_obs))?time_obs:"n/a"):"n/a");
  io.msg(IOL_DFLT,"file version: %.1f\n",version);
  io.msg(IOL_DFLT,"channels: %d\n",nc);
  io.msg(IOL_DFLT,"size: %dx%d\n",abs(cl_xh[0]-cl_xl[0])+1,abs(cl_yh[0]-cl_yl[0])+1);
  io.msg(IOL_DFLT,"subfields: %dx%d\n",nsx,nsy);
  io.msg(IOL_DFLT,"modes: %d\n",nm);
}


void autocorrelate(complex_t **&f,int np,io_class &io)              // 2D autocorrelation of f
{                                                      // general: no pre/post fft shift necessary as the multiply is local
  int nq=2*np;
  for(int x=1;x<=np;++x) memset(f[x]+np+1,0,np*sizeof(complex_t));  // set exterior to 0
  for(int x=np+1;x<=nq;++x) memset(f[x]+1,0,nq*sizeof(complex_t));
  if(fft_2d(f,nq,nq,1)<0) io.msg(IOL_WARN,"Error: autocorrelate: fft_2d failed\n"); // Fourier transform
  for(int x=1;x<=nq;++x)
    for(int y=1;y<=nq;++y){
      f[x][y].re=f[x][y].re*f[x][y].re+f[x][y].im*f[x][y].im;       // Multiply transforms.
      f[x][y].im=0.0;                                               // f*f^* is real
    }
  if(fft_2d(f,nq,nq,-1)<0) io.msg(IOL_WARN,"Error: autocorrelate: fft_2d failed\n");                                               // inverse transform
  fft_reorder(f,nq);
} 

complex_t **wf2psf(float64_t **wf,float32_t **pupil,int nph,int np,io_class &io)
{
//
  complex_t **sj=ct2dim(1,2*nph,1,2*nph);
  memset(sj[1]+1,0,4*nph*nph*sizeof(complex_t));
  float64_t area=0.0;
  for(int u=1;u<=nph;++u)
    for(int v=1;v<=nph;++v){
      sj[u][v].re=pupil[u][v]*cos(wf[u][v]);
      sj[u][v].im=pupil[u][v]*sin(wf[u][v]);
      area+=pupil[u][v];
    }
  autocorrelate(sj,nph,io);
  for(int u=1;u<=2*nph;++u)
    for(int v=1;v<=2*nph;++v){
      sj[u][v].re/=area;
      sj[u][v].im/=area;
    }
// this is incomplete: must also apply the pixel transfer function!
  complex_t **f=ct2dim(1,np,1,np);
  int offs=nph-np/2;
  if(offs>=0){  // data is sampled coarser than the wavefront: this should be done more elegantly, by extending to 2*np, and then binning down to np...
    for(int x=1;x<=np;++x) memcpy(f[x]+1,sj[x+offs]+1+offs,np*sizeof(complex_t));
  }else{ // data is sampled finer than the wavefront
    memset(f[1]+1,0,np*np*sizeof(complex_t));
    for(int x=1;x<=2*nph;++x) memcpy(f[x-offs]+1-offs,sj[x]+1,2*nph*sizeof(complex_t));
  }
  del_ct2dim(sj,1,2*nph,1,2*nph);
//
  fft_reorder(f,np);
  if(fft_2d(f,np,np,-1)<0) io.msg(IOL_WARN,"Error: subimg::psf: fft_2d failed\n"); // inverse transform
  fft_reorder(f,np);                              // psf
//
  return f;
}

void momfbd_struct::centerpatch(int xc,int yc,int &sx,int &sy)
{
  float64_t ix=((float64_t)xc-xo)/sf,iy=((float64_t)yc-yo)/sf;
// locate the patch best centered on the coordinate
  sx=1;
  int metric=sqr(2.0*ix-(float64_t)(patch[1][1]->xh+patch[1][1]->xl));
  for(int x=2;x<=nsx;++x)
    if(metric>sqr(2.0*ix-(float64_t)(patch[x][1]->xh+patch[x][1]->xl))){
      metric=sqr(2.0*ix-(float64_t)(patch[x][1]->xh+patch[x][1]->xl));
      sx=x;
    }
  sy=1;
  metric=sqr(2.0*iy-(float64_t)(patch[sx][1]->yh+patch[sx][1]->yl));
  for(int y=2;y<=nsy;++y)
    if(metric>sqr(2.0*iy-(float64_t)(patch[sx][y]->yh+patch[sx][y]->yl))){
      metric=sqr(2.0*iy-(float64_t)(patch[sx][y]->yh+patch[sx][y]->yl));
      sy=y;
    }
}

//
// generate scaled, averaged PSF for given coordinate and multiplier values. scale must be set at read time for now...
//
float64_t **momfbd_struct::psf(int ix_in,int iy_in,int idx,int mul,float64_t &dx,float64_t &dy,int &np_o,io_class &io)
{
  np_o=2*(int)round(0.5*sf*(float64_t)np); // this formulation assumes that the data scale equals the restored scale!
// convert coordinates to "image" coordinates
  float64_t ix=(((float64_t)ix_in)-xo)/sf,iy=(((float64_t)iy_in)-yo)/sf;
// locate the patch best centered on the coordinate
  int sx=1,sy=1,metric=sqr(ix-patch[1][1]->xc())+sqr(iy-patch[1][1]->yc());
  for(int x=1;x<=nsx;++x)
    for(int y=1;y<=nsy;++y)
      if(metric>(sqr(ix-patch[x][y]->xc())+sqr(iy-patch[x][y]->yc()))){
        metric=sqr(ix-patch[x][y]->xc())+sqr(iy-patch[x][y]->yc());
        sx=x;
        sy=y;
      }
//
  int sxl=sx-(patch[sx][sy]->xc()>ix),syl=sy-(patch[sx][sy]->yc()>iy);
  int sxh=min(max(sxl+1,1),nsx),syh=min(max(syl+1,1),nsy);
  sxl=min(max(sxl,1),nsx);
  syl=min(max(syl,1),nsy);  
// patch center coordinates
  float64_t xcll=patch[sxl][syl]->xc(),xchl=patch[sxh][syl]->xc();
  float64_t xclh=patch[sxl][syh]->xc(),xchh=patch[sxh][syh]->xc();
  float64_t ycll=patch[sxl][syl]->yc(),yclh=patch[sxl][syh]->yc();
  float64_t ychl=patch[sxh][syl]->yc(),ychh=patch[sxh][syh]->yc();
// calculate interpolation factors
  float64_t pxeps=0.1; // margin used to determine if the lower and upper bounds are the same patch or not?
  float64_t fxl=(fabs(xcll-xchl)>pxeps)?(ix-xchl)/(xcll-xchl):0.0;
  float64_t fxh=(fabs(xclh-xchh)>pxeps)?(ix-xchh)/(xclh-xchh):0.0;
  float64_t fy=((fabs(ycll-yclh)>pxeps)&&(fabs(ychl-ychh)>pxeps))?0.5*((iy-yclh)/(ycll-yclh)+(iy-ychh)/(ychl-ychh)):0.0;
// generate the wavefront
  float64_t ***wfll=new float64_t** [mul]-idx,***wflh=new float64_t** [mul]-idx,***wfhl=new float64_t** [mul]-idx,***wfhh=new float64_t** [mul]-idx;
// wavefront tip-tilts
  float64_t *dxll=new float64_t [mul]-idx,*dxlh=new float64_t [mul]-idx,*dxhl=new float64_t [mul]-idx,*dxhh=new float64_t [mul]-idx;
  float64_t *dyll=new float64_t [mul]-idx,*dylh=new float64_t [mul]-idx,*dyhl=new float64_t [mul]-idx,*dyhh=new float64_t [mul]-idx;
// we could save some work here by checking for fxl and fxh as well.
  float64_t dxrf=0.0,dyrf=0.0;
  for(int k=idx;k<=idx+mul-1;++k){
    dxll[k]=dxlh[k]=dxhl[k]=dxhh[k]=dx/sf;
    dyll[k]=dylh[k]=dyhl[k]=dyhh[k]=dy/sf;

    wfll[k]=(fy)?patch[sxl][syl]->wavefront(k,mode,dxll[k],dyll[k],cf2pix):0;
    wflh[k]=patch[sxl][syh]->wavefront(k,mode,dxlh[k],dylh[k],cf2pix);
    wfhl[k]=(fy)?patch[sxh][syl]->wavefront(k,mode,dxhl[k],dyhl[k],cf2pix):0;
    wfhh[k]=patch[sxh][syh]->wavefront(k,mode,dxhh[k],dyhh[k],cf2pix);
// interpolate total tip-tilt to requested location...
    dxrf+=fy*(fxl*dxll[k]+(1.0-fxl)*dxhl[k])+(1.0-fy)*(fxh*dxlh[k]+(1.0-fxh)*dxhh[k]);
    dyrf+=fy*(fxl*dyll[k]+(1.0-fxl)*dyhl[k])+(1.0-fy)*(fxh*dylh[k]+(1.0-fxh)*dyhh[k]);
  }
// calculate optimal average shift in data pixels: this needs to change if pixel scales of the data and restored object differ
  dxrf=round((dxrf/(float64_t)mul)*sf);
  dyrf=round((dyrf/(float64_t)mul)*sf);
// save optimal average shift in pixels
  dx=dxrf;
  dy=dyrf;
//
// average psf over mul frames, add residual tip-tilt w.r.t. optimum integer pixel shift
//
  float64_t **p=f64t2dim(1,np_o,1,np_o);
  memset(p[1]+1,0,np_o*np_o*sizeof(float64_t));
  for(int k=idx;k<=idx+mul-1;++k){
    for(int u=1;u<=nph;++u)
      for(int v=1;v<=nph;++v){ // this must be checked carefully for sign errors!
// add remaining tilt in x-direction
        if(fy) wfll[k][u][v]-=mode[1][u][v]*(dxll[k]-(dxrf/sf))*pix2cf;
        wflh[k][u][v]-=mode[1][u][v]*(dxlh[k]-(dxrf/sf))*pix2cf;
        if(fy) wfhl[k][u][v]-=mode[1][u][v]*(dxhl[k]-(dxrf/sf))*pix2cf;
        wfhh[k][u][v]-=mode[1][u][v]*(dxhh[k]-(dxrf/sf))*pix2cf;
// add remaining tilt in y-direction
        if(fy) wfll[k][u][v]-=mode[2][u][v]*(dyll[k]-(dyrf/sf))*pix2cf;
        wflh[k][u][v]-=mode[2][u][v]*(dylh[k]-(dyrf/sf))*pix2cf;
        if(fy) wfhl[k][u][v]-=mode[2][u][v]*(dyhl[k]-(dyrf/sf))*pix2cf;
        wfhh[k][u][v]-=mode[2][u][v]*(dyhh[k]-(dyrf/sf))*pix2cf;
      }
// calculate psfs
    complex_t **fll=(fy)?wf2psf(wfll[k],pupil,nph,np_o,io):0;
    if(wfll[k]) del_f64t2dim(wfll[k],1,nph,1,nph);
    complex_t **flh=wf2psf(wflh[k],pupil,nph,np_o,io);
    if(wflh[k]) del_f64t2dim(wflh[k],1,nph,1,nph);
    complex_t **fhl=(fy)?wf2psf(wfhl[k],pupil,nph,np_o,io):0;
    if(wfhl[k]) del_f64t2dim(wfhl[k],1,nph,1,nph);
    complex_t **fhh=wf2psf(wfhh[k],pupil,nph,np_o,io);
    if(wfhh[k]) del_f64t2dim(wfhh[k],1,nph,1,nph);
// interpolate psf and add
    if(fy){
      for(int x=1;x<=np_o;++x)
        for(int y=1;y<=np_o;++y) p[x][y]+=fy*(fxl*fll[x][y].re+(1.0-fxl)*fhl[x][y].re)+(1.0-fy)*(fxh*flh[x][y].re+(1.0-fxh)*fhh[x][y].re);
    }else{
      for(int x=1;x<=np_o;++x)
        for(int y=1;y<=np_o;++y) p[x][y]+=fxh*flh[x][y].re+(1.0-fxh)*fhh[x][y].re;
    }
    if(fll) del_ct2dim(fll,1,np_o,1,np_o);
    if(flh) del_ct2dim(flh,1,np_o,1,np_o);
    if(fhl) del_ct2dim(fhl,1,np_o,1,np_o);
    if(fhh) del_ct2dim(fhh,1,np_o,1,np_o);
  }
// minor cleanup...
  delete[] (wfll+idx);
  delete[] (wflh+idx);
  delete[] (wfhl+idx);
  delete[] (wfhh+idx);
  delete[] (dxll+idx);
  delete[] (dxlh+idx);
  delete[] (dxhl+idx);
  delete[] (dxhh+idx);
  delete[] (dyll+idx);
  delete[] (dylh+idx);
  delete[] (dyhl+idx);
  delete[] (dyhh+idx);
// renormalize psf
  for(int x=1;x<=np_o;++x) for(int y=1;y<=np_o;++y) p[x][y]/=(float64_t)mul;
  float64_t sum=0.0;
  for(int x=1;x<=np_o;++x) for(int y=1;y<=np_o;++y) sum+=p[x][y];
  for(int x=1;x<=np_o;++x) for(int y=1;y<=np_o;++y) p[x][y]/=sum;
  if(sum==0.0) io.msg(IOL_WARN,"momfbd_struct::psf: total=%.4lf (ll=%lf lh=%lf hl=%lf hh=%lf %lf %lf %lf)\n",sum,ycll,yclh,ychl,yclh,fy,fxl,fxh);
//  if(sum==0.0) io.msg(IOL_WARN,"momfbd_struct::psf: total=%.4lf (ll=%d lh=%d hl=%d hh=%d %lf %lf %lf)\n",sum,sxl,sxh,syl,syh,fy,fxl,fxh);
//
  return p;
}

float64_t *momfbd_struct::ds(momfbd_struct *that,int di,int sx,int sy,io_class &io)
{
  float64_t *rv=new float64_t [2]-1;
  int ncf;
  float64_t **dalpha=patch[sx][sy]->diff(that->patch[sx][sy],di,ncf);
  float64_t *mdalpha=new float64_t [nm]-1;
  memset(mdalpha+1,0,nm*sizeof(float64_t));
  for(int a=1;a<=ncf;++a) for(int m=1;m<=nm;++m) mdalpha[m]+=dalpha[a][m];
  float64_t **wf=f64t2dim(1,nph,1,nph);
  memset(wf[1]+1,0,nph*nph*sizeof(float64_t));
  for(int m=1;m<=nm;++m) for(int u=1;u<=nph;++u) for(int v=1;v<=nph;++v) wf[u][v]+=mdalpha[m]*mode[m][u][v];
  delete[] (mdalpha+1);
  for(int u=1;u<=nph;++u) for(int v=1;v<=nph;++v) wf[u][v]*=pupil[u][v]/(float64_t)ncf;
  float64_t sumn=0.0,sumd=0.0;
  for(int u=1;u<=nph;++u) 
    for(int v=1;v<=nph;++v){
      sumn+=wf[u][v]*mode[1][u][v];
      sumd+=mode[1][u][v]*mode[1][u][v]*pupil[u][v];
    }
  rv[1]=cf2pix*sumn/sumd;
  sumn=0.0;
  sumd=0.0;
  for(int u=1;u<=nph;++u) 
    for(int v=1;v<=nph;++v){
      sumn+=wf[u][v]*mode[2][u][v];
      sumd+=mode[2][u][v]*mode[2][u][v]*pupil[u][v];
    }
  rv[2]=cf2pix*sumn/sumd;
  del_f64t2dim(wf,1,nph,1,nph);
  del_f64t2dim(dalpha,1,ncf,1,nm);
//
  return rv;
}

/*
float64_t *momfbd_struct::ds(momfbd_struct *that,int di,int sx,int sy,io_class &io)
{
  int ncf;
  float64_t **dalpha=patch[sx][sy]->diff(that->patch[sx][sy],di,ncf);
  float64_t dxc=0.0,dyc=0.0;
  for(int a=1;a<=ncf;++a){
    dxc+=cf2pix*dalpha[a][1];
    dyc+=cf2pix*dalpha[a][2];
  }
  del_f64t2dim(dalpha,1,ncf,1,nm);
//
  float64_t *rv=new float64_t [2]-1;
  rv[1]=dxc/(float64_t)ncf;
  rv[2]=dyc/(float64_t)ncf;
  return rv;
}
*/

float64_t **momfbd_struct::subimg(int &npts,int sx,int sy,io_class &io)
{
  float32_t **tmp=patch[sx][sy]->subimg(npts);
  float64_t **img=f64t2dim(1,npts,1,npts);
  for(int x=1;x<=npts;++x) for(int y=1;y<=npts;++y) img[x][y]=tmp[x][y];
  return img;  
}

float64_t momfbd_struct::imgval(int ix_in,int iy_in,io_class &io)
{
// convert coordinates to "image" coordinates
  float64_t ix=(((float64_t)ix_in)-xo)/sf,iy=(((float64_t)iy_in)-yo)/sf;
// locate the patch best centered on the coordinate
  int sx=1,sy=1,metric=sqr(ix-patch[1][1]->xc())+sqr(iy-patch[1][1]->yc());
  for(int x=1;x<=nsx;++x)
    for(int y=1;y<=nsy;++y)
      if(metric>(sqr(ix-patch[x][y]->xc())+sqr(iy-patch[x][y]->yc()))){
        metric=sqr(ix-patch[x][y]->xc())+sqr(iy-patch[x][y]->yc());
        sx=x;
        sy=y;
      }
//
  int sxl=sx-(patch[sx][sy]->xc()>ix),syl=sy-(patch[sx][sy]->yc()>iy);
//
  if((sxl<1)||(sxl>=nsx)||(syl<1)||(syl>=nsy)) return 0.0; // make sure we can see what is really covered by the MOMFBD file
//
  int sxh=min(max(sxl+1,1),nsx),syh=min(max(syl+1,1),nsy);
//
  sxl=min(max(sxl,1),nsx);
  syl=min(max(syl,1),nsy);  
// patch center coordinates
  float64_t xcll=patch[sxl][syl]->xc(),xchl=patch[sxh][syl]->xc();
  float64_t xclh=patch[sxl][syh]->xc(),xchh=patch[sxh][syh]->xc();
  float64_t ycll=patch[sxl][syl]->yc(),yclh=patch[sxl][syh]->yc();
  float64_t ychl=patch[sxh][syl]->yc(),ychh=patch[sxh][syh]->yc();
// calculate interpolation factors
  float64_t pxeps=0.1; // margin used to determine if the lower and upper bounds are the same patch or not?
  float64_t fxl=(fabs(xcll-xchl)>pxeps)?(ix-xchl)/(xcll-xchl):0.0;
  float64_t fxh=(fabs(xclh-xchh)>pxeps)?(ix-xchh)/(xclh-xchh):0.0;
  float64_t fy=((fabs(ycll-yclh)>pxeps)&&(fabs(ychl-ychh)>pxeps))?0.5*((iy-yclh)/(ycll-yclh)+(iy-ychh)/(ychl-ychh)):0.0;
//
  float64_t fll=(fy)?patch[sxl][syl]->imgval(ix,iy):0;
  float64_t flh=patch[sxl][syh]->imgval(ix,iy);
  float64_t fhl=(fy)?patch[sxh][syl]->imgval(ix,iy):0;
  float64_t fhh=patch[sxh][syh]->imgval(ix,iy);
//
  float64_t fl=((fll>-1.0)&&(fhl>-1.0))?fxl*fll+(1.0-fxl)*fhl:(fll>-1.0)?fll:fhl;
  float64_t fh=((flh>-1.0)&&(fhh>-1.0))?fxh*flh+(1.0-fxh)*fhh:(flh>-1.0)?flh:fhh;
  return ((fl>-1.0)&&(fh>-1.0))?fy*fl+(1.0-fy)*fh:(fl>-1.0)?fl:fh;
  
  
//  return fy*(fxl*fll+(1.0-fxl)*fhl)+(1.0-fy)*(fxh*flh+(1.0-fxh)*fhh);
//  return fy*(fxl*fll+(1.0-fxl)*fhl)+(1.0-fy)*(fxh*flh+(1.0-fxh)*fhh);
//  return (fy>0.5)?((fxl>0.5)?fll:fhl):((fxh>0.5)?flh:fhh);
}

char **momfbd_append(char **a,const char *b)
{
  int la=0;
  if(a) while(a[la]) ++la;
//
  char **tmp=new char* [la+2];
//
  if(la) memcpy(tmp,a,la*sizeof(char*));
  tmp[la]=strcpy(new char [strlen(b)+1],b);
  tmp[la+1]=0;
//
  return tmp;
}


char **momfbd_struct::fetchdata(char **dll,int mul,char *rawpath,io_class &io)
{
  for(int i=1;i<=patch[1][1]->nim[1];i+=mul){
    struct stat stat_buf;
    char path[1000];
    if(char *q=strrchr(names[i-1],'/')){
      sprintf(path,"%s/%s",rawpath,q+1);
      if(stat(path,&stat_buf)<0){
        int found=0;
        for(int k=0;dll[k];++k) // remove duplicates
          if(!strcmp(dll[k],q+1)){
            found=1;
            break;
          }
        if(!found) dll=momfbd_append(dll,strrchr(names[i-1],'/')+1); // file not found!
      }
    }
  }
  return dll;
}

char *momfbd_struct::fetchdatapath(io_class &io)
{
  if(char *q=strrchr(names[0],'/')){
    char *path=new char [strlen(names[0])+1];
    memcpy(path,names[0],q-names[0]);
    path[q-names[0]]=0;
    return path;
  }
  return 0;
}

#define PI 3.141592653589793238462643383279502884197169399375105

static __inline fp_t sf(fp_t x) // x=0..1
{
  return 0.5*(1.0-cos(PI*x));
}

void subimg_paste(float32_t **pic,int nx,int ny,float32_t **sub_img,int x,int y,int16_t *xl,int16_t *xh,int16_t *yl,int16_t *yh,int mx,int my,io_class &io)
{
  int snx=xh[x]-xl[x]+1,sny=yh[y]-yl[y]+1;
  float32_t **sim=sub_img;
  int margin=8;
  int offx=(xh[x]-xl[x]+1)/margin,offy=(yh[y]-yl[y]+1)/margin;
  int exl=xl[x]+offx,exh=xh[x]-offx,eyl=yl[y]+offy,eyh=yh[y]-offy;
  if((exl<1)||(eyl<1)||(exh>nx)||(eyh>ny)){
    io.msg(IOL_WARN,"cannot paste subimage (%d,%d,%d,%d) into image (1,%d,1,%d). clipping...\n",xl[x],xh[x],yl[y],yh[y],nx,ny);
    if(exl<1) exl=1;
    if(eyl<1) eyl=1;
    if(exh>nx) exh=nx;
    if(eyh>ny) eyh=ny;
  }
  int ixl=(x> 1)?xh[x-1]-(xh[x-1]-xl[x-1]+1)/margin+1:exl;
  int ixh=(x<mx)?xl[x+1]+(xh[x+1]-xl[x+1]+1)/margin-1:exh;
  int iyl=(y> 1)?yh[y-1]-(yh[y-1]-yl[y-1]+1)/margin+1:eyl;
  int iyh=(y<my)?yl[y+1]+(yh[y+1]-yl[y+1]+1)/margin-1:eyh;
//
  float32_t **window=f32t2dim(exl,exh,eyl,eyh);
  io.msg(IOL_INFO,"%d %d %d %d\n",exl,exh,eyl,eyh);
  for(int ix=exl;ix<=exh;++ix)
    for(int iy=eyl;iy<=eyh;++iy) window[ix][iy]=1.0;
  for(int ix=exl;ix<=ixl-1;++ix){ // lower x margin
    float32_t xx=(float32_t)(ix-exl+1)/(float32_t)(ixl-exl+1);
    for(int iy=eyl;iy<=eyh;++iy){
      if((ix<exl)||(ix>exh)) io.msg(IOL_INFO,"in x: %d %d\n",ix,iy);
      if((iy<eyl)||(iy>eyh)) io.msg(IOL_INFO,"in y: %d %d\n",ix,iy);
      window[ix][iy]*=sf(xx)/(sf(xx)+sf(1.0-xx));
    }
  }
  for(int ix=ixh+1;ix<=exh;++ix){ // upper x margin
    float32_t xx=1.0-(float32_t)(ix-ixh)/(float32_t)(exh-ixh+1);
    for(int iy=eyl;iy<=eyh;++iy) window[ix][iy]*=sf(xx)/(sf(xx)+sf(1.0-xx));
  }
  for(int iy=eyl;iy<=iyl-1;++iy){ // lower y margin
    float32_t yy=(float32_t)(iy-eyl+1)/(float32_t)(iyl-eyl+1);
    for(int ix=exl;ix<=exh;++ix) window[ix][iy]*=sf(yy)/(sf(yy)+sf(1.0-yy));
  }
  for(int iy=iyh+1;iy<=eyh;++iy){ // upper y margin
    float32_t yy=1.0-(float32_t)(iy-iyh)/(float32_t)(eyh-iyh+1);
    for(int ix=exl;ix<=exh;++ix) window[ix][iy]*=sf(yy)/(sf(yy)+sf(1.0-yy));
  }
  for(int ix=exl;ix<=exh;++ix)
    for(int iy=eyl;iy<=eyh;++iy)
      pic[ix][iy]+=(float32_t)(window[ix][iy]*sim[ix-xl[x]+1][iy-yl[y]+1]);
  del_f32t2dim(window,exl,exh,eyl,eyh);
}

void img_clip(float32_t **pic,int &nx,int &ny)
{
  float32_t *xs=new float32_t [nx] -1,*ys=new float32_t [ny] -1; 
  memset(xs+1,0,nx*sizeof(float32_t));
  memset(ys+1,0,ny*sizeof(float32_t));
  for(int x=1;x<=nx;++x)
    for(int y=1;y<=ny;++y){
      xs[x]+=pic[x][y];
      ys[y]+=pic[x][y];
    }
  int xl,xh,yl,yh;
  for(xl=1;xs[xl]==0;++xl);
  for(xh=nx;xs[xh]==0;--xh);
  for(yl=1;ys[yl]==0;++yl);
  for(yh=ny;ys[yh]==0;--yh);
  delete[] (xs+1);
  delete[] (ys+1);
  nx=xh-xl+1;
  ny=yh-yl+1;
  for(int x=xl;x<=xh;++x) memmove(pic[1]+(x-xl)*ny+1,pic[x]+yl,ny*sizeof(float32_t));
  for(int x=2;x<=nx;++x) pic[x]=pic[x-1]+ny; // update pointers
}

/*
byte *momfbd_read(char *file_name,int *&ds,int &nd,char *&header,int &type,io_class &io) // fzread subroutine	
{
  if(int fd=open(file_name,O_RDONLY)){
    int one=1;
    uint08_t endian=(*(char*)&one==0); // 1=LITTLE_ENDIAN, 0=BIG_ENDIAN: taken from SL's tiff library 
    uint08_t f_endian;
    read(fd,&f_endian,1);
    uint08_t swap_endian=(f_endian!=endian);
//
    io.msg(IOL_INFO,"byte swapping: %s\n",(swap_endian)?"yes":"no");
//
    int32_t strsz;
    read(fd,&strsz,sizeof(int32_t));
    if(swap_endian) bswap(&strsz,1);
    char *h=new char [strsz];
    read(fd,h,strsz);
    io.msg(IOL_INFO,"header: %s\n",h);
//
    read(fd,&strsz,sizeof(int32_t));
    if(swap_endian) bswap(&strsz,1);
    char *time_obs=new char [strsz];
    read(fd,time_obs,strsz);
    io.msg(IOL_INFO,"time: %s\n",time_obs);
//
    read(fd,&strsz,sizeof(int32_t));
    if(swap_endian) bswap(&strsz,1);
    char *date_obs=new char [strsz];
    read(fd,date_obs,strsz);
    io.msg(IOL_INFO,"date: %s\n",date_obs);
//
    uint08_t has_modes;
    read(fd,&has_modes,1);
    io.msg(IOL_INFO,"modes: %s\n",(has_modes)?"yes":"no");
    float32_t ***mode=0;
    float32_t **pupil=0;
    int32_t nph,nm;
    float32_t pix2cf=NAN,cf2pix=NAN;
    float32_t version=0.0;
    sscanf(h,"%f",&version);
    if(has_modes){
      if(version>=20110714.0){             // check for version!?
        read(fd,&pix2cf,sizeof(float32_t));
        read(fd,&cf2pix,sizeof(float32_t));
        if(swap_endian){
          bswap(&pix2cf,1);
          bswap(&cf2pix,1);
        }
      }
      read(fd,&nph,sizeof(int32_t));
      read(fd,&nm,sizeof(int32_t));
      if(swap_endian){
        bswap(&nph,1);
        bswap(&nm,1);
      }
      pupil=f32t2dim(1,nph,1,nph);
      read(fd,pupil[1]+1,nph*nph*sizeof(float32_t));
      if(swap_endian) bswap(pupil[1]+1,nph*nph);

      io.msg(IOL_INFO,"reading %dx%dx%d ... ",nm,nph,nph);
      mode=f32t3dim(1,nm,1,nph,1,nph);
      for(int m=1;m<=nm;++m){
        int08_t typ;
        if(version>=20110916.0) read(fd,&typ,1); // mode type
        read(fd,mode[m][1]+1,nph*nph*sizeof(float32_t));
      }
//      read(fd,mode[1][1]+1,nm*nph*nph*sizeof(float32_t));
      if(swap_endian) bswap(mode[1][1]+1,nm*nph*nph);
      io.msg(IOL_INFO,"done.\n");
    }
// globals
    int32_t nc;
    read(fd,&nc,sizeof(int32_t));
    if(swap_endian) bswap(&nc,1);
    io.msg(IOL_INFO,"nc=%d\n",nc);
// global clip
    int16_t cl_xl[nc],cl_xh[nc],cl_yl[nc],cl_yh[nc];
    if(read(fd,cl_xl,nc*sizeof(int16_t))<nc*sizeof(int16_t)) io.msg(IOL_WARN,"short read: \n",strerror(errno));
    if(read(fd,cl_xh,nc*sizeof(int16_t))<nc*sizeof(int16_t)) io.msg(IOL_WARN,"short read: \n",strerror(errno));
    if(read(fd,cl_yl,nc*sizeof(int16_t))<nc*sizeof(int16_t)) io.msg(IOL_WARN,"short read: \n",strerror(errno));
    if(read(fd,cl_yh,nc*sizeof(int16_t))<nc*sizeof(int16_t)) io.msg(IOL_WARN,"short read: \n",strerror(errno));
    if(swap_endian){
      bswap(cl_xl,nc);
      bswap(cl_xh,nc);
      bswap(cl_yl,nc);
      bswap(cl_yh,nc);
    }
//
    for(int c=0;c<nc;++c) io.msg(IOL_XNFO,"clip=(%d,%d,%d,%d)\n",cl_xl[c],cl_xh[c],cl_yl[c],cl_yh[c]);
//
    int32_t nsx,nsy,np;
    read(fd,&nsx,sizeof(int32_t));
    read(fd,&nsy,sizeof(int32_t));
    read(fd,&np,sizeof(int32_t));
    if(swap_endian){
      bswap(&nsx,1);
      bswap(&nsy,1);
      bswap(&np,1);
    }
//
    io.msg(IOL_INFO,"reading %d x %d\n",nsx,nsy);
//
    patch_struct ***data=(patch_struct***)v2dim(1,nsx,1,nsy);
    for(int x=1;x<=nsx;++x)
      for(int y=1;y<=nsy;++y) data[x][y]=new patch_struct(fd,swap_endian,h);
    close(fd);
//
    int16_t *xl=new int16_t [nsx]-1,*xh=new int16_t [nsx]-1;
    for(int x=1;x<=nsx;++x){
      xl[x]=data[x][1]->xl;
      xh[x]=data[x][1]->xh;
    } 
    int16_t *yl=new int16_t [nsx]-1,*yh=new int16_t [nsx]-1;
    for(int y=1;y<=nsy;++y){
      yl[y]=data[1][y]->yl;
      yh[y]=data[1][y]->yh;
    } 
//
    int XL=data[1][1]->xl,XH=data[1][1]->xh;
    for(int x=2;x<=nsx;++x){
      XL=min(data[x][1]->xl,XL);
      XH=max(data[x][1]->xh,XH);
    }
    int YL=data[1][1]->yl,YH=data[1][1]->yh;
    for(int y=2;y<=nsy;++y){
      YL=min(data[1][y]->yl,YL);
      YH=max(data[1][y]->yh,YH);
    }
//
//    int NX=XH-XL+1;
//    int NY=YH-YL+1;
    int NX=(cl_xh[0]>cl_xl[0])?cl_xh[0]-cl_xl[0]+1:cl_xl[0]-cl_xh[0]+1;
    int NY=(cl_yh[0]>cl_yl[0])?cl_yh[0]-cl_yl[0]+1:cl_yl[0]-cl_yh[0]+1;
    float32_t **pic=f32t2dim(1,NX,1,NY);
    for(int x=1;x<=nsx;++x)
      for(int y=1;y<=nsy;++y){
        subimg_paste(pic,NX,NY,data[x][y]->img,x,y,xl,xh,yl,yh,nsx,nsy,io);
        delete data[x][y];
      }
    del_v2dim((void***)data,1,nsx,1,nsy);
//
//    img_clip(pic,NX,NY);
//    in_place_transpose(pic[1]+1,NY,NX);
//
    delete[] (xl+1);
    delete[] (xh+1);
    delete[] (yl+1);
    delete[] (yh+1);
//
    byte *p=(byte*)(pic[1]+1);
    delete[] (pic+1);
    nd=2;
    ds=new int [nd]-1;
    ds[1]=NX;
    ds[2]=NY;
    type=MOMFBD_FLOAT32;
    header=new char [strlen(time_obs)+strlen(time_obs)+1000];
    sprintf(header,"TIME_OBS=%s  DATE_OBS=%s",time_obs,date_obs);
    delete[] date_obs;
    delete[] time_obs;
    return p;
  }
  return 0;
}
*/
